@@ -2,23 +2,24 @@
  *  
  *   syscalls.c: syscall mgmt
  *   
- *   Copyright 2006 Renzo Davoli University of Bologna - Italy
+ *   Copyright 2006-2017 Renzo Davoli University of Bologna - Italy
  *   Copyright 2005 Andrea Gasparini University of Bologna - Italy
  *   
- *   This program is free software; you can redistribute it and/or modify
- *   it under the terms of the GNU General Public License as published by
- *   the Free Software Foundation; either version 2 of the License, or
- *   (at your option) any later version.
- *
- *   This program is distributed in the hope that it will be useful,
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details.
- *
- *   You should have received a copy of the GNU General Public License a
- *   with this program; if not, write to the Free Software Foundation, Inc.,
- *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- */   
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ * 
+ * The GNU C Library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with the GNU C Library; if not, write to the Free
+ * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301, USA.
+ */ 
 
 #include <config.h>
 #include <stdarg.h>
@@ -43,6 +44,7 @@
 #include <sys/xattr.h>
 #include <sys/timeb.h>
 #include <sys/mman.h>
+#include <sys/epoll.h>
 #include <bits/wordsize.h>
 #include <utime.h>
 #include <stdarg.h>
@@ -51,7 +53,6 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <dlfcn.h>
-#include <ustat.h>
 #include <string.h>
 #include <time.h>
 #include <grp.h>
@@ -59,10 +60,11 @@
 #include <stdlib.h>
 #include "purelibc.h"
 
-sfun _pure_syscall=syscall;
-sfun _pure_native_syscall=syscall;
-extern sfun _pure_socketcall;
+/* in case of pre-init call, use syscall */
+sfun _pure_syscall = syscall;
+sfun _pure_native_syscall;
 
+#if __WORDSIZE != 64
 static int _pageshift()
 {
 	static int ps=0;
@@ -73,6 +75,7 @@ static int _pageshift()
 	}
 	return ps;
 }
+#endif
 
 long _pure_debug_printf(const char *format, ...)
 {
@@ -106,6 +109,10 @@ int open(const char* pathname,int flags,...){
 		return _pure_syscall(__NR_open,pathname,flags);
 }
 
+int __open_2(const char* pathname,int flags){
+	return _pure_syscall(__NR_open,pathname,flags);
+}
+
 int open64(const char* pathname,int flags,...){
 	va_list arg_list;
 	if( flags &  O_CREAT ){
@@ -119,6 +126,10 @@ int open64(const char* pathname,int flags,...){
 		return _pure_syscall(__NR_open,pathname,flags|O_LARGEFILE);
 }
 
+int __open64_2 (const char* pathname,int flags){
+	return _pure_syscall(__NR_open,pathname,flags|O_LARGEFILE);
+}
+
 int creat(const char *pathname, mode_t mode)
 {
 	return _pure_syscall(__NR_open,pathname,O_CREAT|O_WRONLY|O_TRUNC,mode);
@@ -222,7 +233,7 @@ int __xstat(int ver, const char* pathname, struct stat* buf_stat)
 			break;
 
 		default:
-			fprintf(stderr, "*** BUG! *** __xstat can't manage version %d!\n", ver);
+			_pure_debug_printf("*** BUG! *** __xstat can't manage version %d!\n", ver);
 			abort();
 	}
 
@@ -244,7 +255,7 @@ int __lxstat(int ver, const char* pathname, struct stat* buf_stat)
 			break;
 
 		default:
-			fprintf(stderr, "*** BUG! *** __lxstat can't manage version %d!\n", ver);
+			_pure_debug_printf("*** BUG! *** __lxstat can't manage version %d!\n", ver);
 			abort();
 	}
 
@@ -265,7 +276,7 @@ int __fxstat(int ver, int fildes, struct stat* buf_stat)
 			break;
 
 		default:
-			fprintf(stderr, "*** BUG! *** __fxstat can't manage version %d!\n", ver);
+			_pure_debug_printf("*** BUG! *** __fxstat can't manage version %d!\n", ver);
 			abort();
 	}
 	if (rv >= 0)
@@ -298,6 +309,7 @@ int __xmknod (int ver, const char *path, mode_t mode, dev_t *dev) {
 int access(const char* pathname,int mode){
 	return _pure_syscall(__NR_access,pathname,mode);
 }
+
 int __access(const char* pathname,int mode){
 	return access(pathname,mode);
 }
@@ -370,14 +382,22 @@ int fchdir(int fd) {
 	return _pure_syscall(__NR_fchdir,fd);
 }
 
-int utime(const char* pathname,const struct utimbuf *buf){
-	return _pure_syscall(__NR_utime,pathname,buf);
-}
-
 int utimes(const char* pathname,const struct timeval tv[2]){
 	return _pure_syscall(__NR_utimes,pathname,tv);
 }
 
+int utime(const char* pathname,const struct utimbuf *buf){
+#ifdef __NR_utime
+	return _pure_syscall(__NR_utime,pathname,buf);
+#else
+	struct timeval tv[2];
+	tv[0].tv_sec = buf->actime;
+	tv[1].tv_sec = buf->modtime;
+	tv[0].tv_usec = tv[1].tv_usec = 0;
+	return utimes(pathname, tv);
+#endif
+}
+
 #ifdef __NR_pread
 ssize_t pread(int fs,void* buf, size_t count, __off_t offset){
 	return _pure_syscall(__NR_pread,fs,buf,count,offset);
@@ -393,7 +413,7 @@ ssize_t pwrite(int fs,const void* buf, size_t count, __off_t offset){
 #ifdef __NR_pread64
 ssize_t pread64(int fs,void* buf, size_t count, __off64_t offset){
 	return _pure_syscall(__NR_pread64,fs,buf,count,
-#if defined(__powerpc__)
+#if defined(__powerpc__) || defined(__arm__)
 			0,
 #endif
 			__LONG_LONG_PAIR( (__off_t)(offset>>32),(__off_t)(offset&0xffffffff)));
@@ -406,7 +426,7 @@ ssize_t pread(int fs,void* buf, size_t count, __off_t offset){
 #ifdef __NR_pwrite64 
 ssize_t pwrite64(int fs,const void* buf, size_t count, __off64_t offset){
 	return _pure_syscall(__NR_pwrite64,fs,buf,count,
-#if defined(__powerpc__)
+#if defined(__powerpc__) || defined(__arm__)
 			0,
 #endif
 			__LONG_LONG_PAIR( (__off_t)(offset>>32),(__off_t)(offset&0xffffffff)));
@@ -420,7 +440,7 @@ ssize_t pwrite(int fs,const void* buf, size_t count, __off_t offset){
 ssize_t preadv64(int fs,const struct iovec *iov, int iovcnt, __off64_t offset){
 	ssize_t rv=_pure_syscall(__NR_preadv,fs,iov,iovcnt,
 #ifdef __NR_pread64
-#if defined(__powerpc__)
+#if defined(__powerpc__) || defined(__arm__)
 			0,
 #endif
 			__LONG_LONG_PAIR( (__off_t)(offset>>32),(__off_t)(offset&0xffffffff))
@@ -459,7 +479,7 @@ ssize_t preadv(int fs,const struct iovec *iov, int iovcnt, __off_t offset){
 ssize_t pwritev64(int fs,const struct iovec *iov, int iovcnt, __off64_t offset){
 	ssize_t rv=_pure_syscall(__NR_pwritev,fs,iov,iovcnt,
 #ifdef __NR_pwrite64
-#if defined(__powerpc__)
+#if defined(__powerpc__) || defined(__arm__)
 			0,    
 #endif
 			__LONG_LONG_PAIR( (__off_t)(offset>>32),(__off_t)(offset&0xffffffff))
@@ -491,18 +511,24 @@ ssize_t pwritev(int fs,const struct iovec *iov, int iovcnt, __off_t offset){
 }
 #endif
 
-int getdents(int fd, long dirp,unsigned int count){
-	return _pure_syscall(__NR_getdents,fd,dirp,count);
+int getdents(int fd, void *dirp, unsigned int count){
+	return _pure_syscall(__NR_getdents, fd, dirp, count);
 }
 
-int getdents64(int fd, long dirp,unsigned int count){
-	return _pure_syscall(__NR_getdents64,fd,dirp,count);
+__ssize_t getdents64(int fd, void *dirp, size_t count){
+	return _pure_syscall(__NR_getdents64, fd, dirp, count);
 }
 
 __off_t lseek(int fd,__off_t offset,int whence){
 	return _pure_syscall(__NR_lseek,fd,offset,whence);
 }
 
+#ifndef __NR__llseek
+off64_t lseek64(int fd, off64_t offset, int whence) {
+	return lseek(fd, offset, whence);
+}
+#endif
+
 #ifdef __NR__llseek
 __off64_t lseek64(int fd, __off64_t offset, int whence){
 	unsigned long offset_high=offset >> (sizeof(unsigned long) * 8);
@@ -786,10 +812,6 @@ pid_t vfork(void){
 	return _pure_syscall(__NR_fork);
 }
 
-time_t time(time_t *t){
-	return _pure_syscall(__NR_time,t);
-}
-
 int stime(const time_t *t){
 	struct timeval tivu = { *t,0};
 	return _pure_syscall(__NR_settimeofday,&tivu,NULL);
@@ -833,6 +855,7 @@ clock_t times(struct tms *buf){
 	return _pure_syscall(__NR_times,buf);
 }
 
+struct ustat;
 int ustat(dev_t dev, struct ustat *ubuf){
 	return _pure_syscall(__NR_ustat,dev,ubuf);
 }
@@ -849,6 +872,20 @@ int sethostname(const char *name, size_t len){
 	return _pure_syscall(__NR_sethostname,name,len);
 }
 
+#ifdef __NR_prlimit64
+int prlimit(pid_t pid, enum __rlimit_resource resource, const struct rlimit *new_limit,
+		struct rlimit *old_limit) {
+	return _pure_syscall(__NR_prlimit64, pid, resource, new_limit, old_limit);
+}
+
+int setrlimit(__rlimit_resource_t resource, const struct rlimit *rlim){
+	return prlimit(0,resource,rlim,NULL);
+}
+
+int getrlimit(__rlimit_resource_t resource, struct rlimit *rlim){ 
+	return prlimit(0,resource,NULL,rlim);
+}
+#else
 int setrlimit(__rlimit_resource_t resource, const struct rlimit *rlim){
 	return _pure_syscall(__NR_setrlimit,resource,rlim);
 }
@@ -856,12 +893,18 @@ int setrlimit(__rlimit_resource_t resource, const struct rlimit *rlim){
 int getrlimit(__rlimit_resource_t resource, struct rlimit *rlim){ 
 	return _pure_syscall(__NR_getrlimit,resource,rlim);
 }
+#endif
 
 int getrusage(int who, struct rusage *usage){
 	return _pure_syscall(__NR_getrusage,usage);
 }
 
-int gettimeofday(struct timeval *tv, struct timezone *tz){
+#ifdef GETTIMEOFDAY_TZ
+int gettimeofday(struct timeval *tv, struct timezone *tz)
+#else
+int gettimeofday(struct timeval *tv, void *tz)
+#endif
+{
 	return _pure_syscall(__NR_gettimeofday, tv, tz);
 }
 
@@ -869,6 +912,19 @@ int settimeofday(const struct timeval *tv , const struct timezone *tz){
 	return _pure_syscall(__NR_settimeofday, tv, tz);
 }
 
+time_t time(time_t *t){
+#ifdef __NR_time
+	return _pure_syscall(__NR_time,t);
+#else
+	struct timeval tv;
+	if (gettimeofday(&tv, NULL) == 0) {
+		if (t) *t = tv.tv_sec;
+		return tv.tv_sec;
+	} else
+		return -1;
+#endif
+}
+
 int getgroups(int size, gid_t list[]){
 	return _pure_syscall(__NR_getgroups,size,list);
 }
@@ -888,6 +944,46 @@ int poll(struct pollfd *ufds, nfds_t nfds, int timeout){
 	return _pure_syscall(__NR_poll,ufds,nfds,timeout);
 }
 
+int pselect(int nfds, fd_set *readfds, fd_set *writefds,
+		fd_set *exceptfds, const struct timespec *timeout,
+		const sigset_t *sigmask) {
+	return _pure_syscall(__NR_pselect6,nfds, readfds, writefds, exceptfds, timeout, sigmask);
+}
+
+int ppoll(struct pollfd *fds, nfds_t nfds,
+		const struct timespec *tmo_p, const sigset_t *sigmask) {
+	return _pure_syscall(__NR_ppoll,fds,nfds,tmo_p,sigmask);
+}
+
+#ifdef __NR_epoll_create1
+int epoll_create1(int flags) {
+	return _pure_syscall(__NR_epoll_create1, flags);
+}
+
+int epoll_create(int size) {
+	if (size <= 0) {
+		errno = EINVAL;
+		return -1;
+	} else
+		return _pure_syscall(__NR_epoll_create1, 0);
+}
+
+int epoll_ctl(int epfd, int op, int fd, struct epoll_event *event) {
+	return _pure_syscall(__NR_epoll_ctl,epfd,op,fd,event);
+}
+
+int epoll_wait(int epfd, struct epoll_event *events,
+		int maxevents, int timeout) {
+	return _pure_syscall(__NR_epoll_pwait,epfd,events,maxevents,timeout,NULL);
+}
+
+int epoll_pwait(int epfd, struct epoll_event *events,
+		int maxevents, int timeout,
+		const sigset_t *sigmask) {
+	return _pure_syscall(__NR_epoll_pwait,epfd,events,maxevents,timeout,sigmask);
+}
+#endif
+
 int truncate(const char *path, __off_t length){
 	return _pure_syscall(__NR_truncate,path,length);
 }
@@ -1245,6 +1341,16 @@ int openat64(int dirfd,const char* pathname,int flags,...){
 	else
 		return _pure_syscall(__NR_openat,dirfd,pathname,flags|O_LARGEFILE);
 }
+
+int	__openat_2(int dirfd, const char *pathname, int flags)
+{
+	return _pure_syscall(__NR_openat,dirfd,pathname,flags);
+}
+
+int	__openat64_2(int dirfd, const char *pathname, int flags)
+{
+	return _pure_syscall(__NR_openat,dirfd,pathname,flags|O_LARGEFILE);
+}
 #endif
 
 #ifdef __NR_mkdirat
@@ -1295,7 +1401,7 @@ int __fxstatat(int ver, int fildes, const char *pathname, struct stat* buf_stat,
 			break;
 
 		default:
-			fprintf(stderr, "*** BUG! *** __fxstatat can't manage version %d!\n", ver);
+			_pure_debug_printf("*** BUG! *** __fxstatat can't manage version %d!\n", ver);
 			abort();
 	}
 	if (rv >= 0)
@@ -1318,7 +1424,7 @@ int renameat(int olddirfd, const char *oldpath,int newdirfd, const char *newpath
 
 #ifdef __NR_linkat
 int linkat(int olddirfd, const char *oldpath,int newdirfd, const char *newpath, int flags){
-	return _pure_syscall(__NR_linkat,oldpath,newdirfd,newpath,flags);
+	return _pure_syscall(__NR_linkat,olddirfd,oldpath,newdirfd,newpath,flags);
 }
 #endif
 
@@ -1330,7 +1436,7 @@ int symlinkat(const char *oldpath, int newdirfd, const char *newpath){
 
 #ifdef __NR_readlinkat
 ssize_t readlinkat(int dirfd, const char *pathname,char *buf, size_t bufsiz){
-	return _pure_syscall(__NR_readlinkat,dirfd,pathname,bufsiz);
+	return _pure_syscall(__NR_readlinkat,dirfd,pathname,buf,bufsiz);
 }
 #endif
 
@@ -1369,7 +1475,8 @@ long int syscall(long int n,...)
 	arg4=va_arg(ap, long int);
 	arg5=va_arg(ap, long int);
 	va_end(ap);
-	if (__builtin_expect(_pure_native_syscall == syscall,0))
+	/* in case of pre-init call: emergency initialization */ 
+	if (__builtin_expect(_pure_native_syscall == NULL,0))
 		init();
 	if (__builtin_expect(_pure_syscall == syscall,0))
 		return _pure_native_syscall(n,arg0,arg1,arg2,arg3,arg4,arg5);
@@ -1377,7 +1484,7 @@ long int syscall(long int n,...)
 		return _pure_syscall(n,arg0,arg1,arg2,arg3,arg4,arg5);
 }
 
-sfun _pure_start(sfun pure_syscall,sfun pure_socketcall,int flags)
+sfun _pure_start(sfun pure_syscall, int flags)
 {
 	int fdtmp;
 	if (__builtin_expect(_pure_native_syscall == syscall,0))
@@ -1409,16 +1516,16 @@ sfun _pure_start(sfun pure_syscall,sfun pure_socketcall,int flags)
 		close(fdtmp);
 	}
 	_pure_syscall=pure_syscall;
-	if (pure_socketcall)
-		_pure_socketcall=pure_socketcall;
 	return _pure_native_syscall;
 }
 
-	static void
+/* this is convenient since casting the return value of dlsym() to
+ * a function pointer erroneously procudes a warning */
+#pragma GCC diagnostic ignored "-Wpedantic"
+
 	__attribute__ ((constructor))
-init (void)
+static void init (void)
 {
-	if (__builtin_expect(_pure_native_syscall == syscall,0))
-		_pure_native_syscall=dlsym(RTLD_NEXT,"syscall");
+	_pure_native_syscall = _pure_syscall = dlsym(RTLD_NEXT,"syscall");
 }
 
